# ------------------------------------------------------------------------------
# ** SplitToPieces **
# Splits body into multiple pieces for user purpose.
#
# Functions adapted from SketchyPhysics by Chris Phillips.
# Written by Anton Synytsia
# ------------------------------------------------------------------------------

require 'sketchup'

module AMS; end
module AMS::SplitToPieces

  VERSION = '1.0.0'
  RELEASE_DATE = 'October 17, 2014'

  class << self

    def split(grp, recurse)
      recurse = recurse.to_i
      recurse = 0 if recurse < 0
      grps = cut(grp)
      if recurse > 0
        grps.concat( split(grps[0], recurse-1) ) if grps[0]
        grps.concat( split(grps[1], recurse-1) ) if grps[1]
      end
      grps.reject! { |e| e.deleted? }
      grps
    end

    def cut(grp)
      ents = Sketchup.active_model.entities

      cut_group = ents.add_group
      xlate = Geom::Transformation.new(grp.bounds.center)
      rot = xlate * Geom::Transformation.rotation(ORIGIN, [rand-0.5, rand-0.5, rand-0.5], (rand-0.5)*10.0)
      d = grp.bounds.diagonal*2
      f = cut_group.entities.add_face(
        [-d, -d, 0].transform!(rot),
        [ d, -d, 0].transform!(rot),
        [ d,  d, 0].transform!(rot),
        [-d,  d, 0].transform!(rot)
      )
      pl = f.plane

      agrp = ents.add_instance(get_definition(grp), grp.transformation)
      agrp.material = grp.material
      agrp.make_unique
      adef = get_definition(agrp)
      edges = adef.entities.intersect_with(false, agrp.transformation, adef.entities, agrp.transformation, false, [cut_group])
      delete_ents_behind_plane(adef.entities, agrp.transformation, pl)
      edges.each { |e| e.find_faces if e.valid? }

      for i in 0..3; pl[i] *= -1; end
      bgrp = ents.add_instance(get_definition(grp), grp.transformation)
      bgrp.material = grp.material
      bgrp.make_unique
      bdef = get_definition(bgrp)
      edges = bdef.entities.intersect_with(false, bgrp.transformation, bdef.entities, bgrp.transformation, false, [cut_group])
      delete_ents_behind_plane(bdef.entities, bgrp.transformation, pl)
      edges.each { |e| e.find_faces if e.valid? }

      cut_group.erase!
      agrp.erase! if agrp.entities.count.zero?
      bgrp.erase! if bgrp.entities.count.zero?

      grps = []
      grps << agrp if agrp.valid?
      grps << bgrp if bgrp.valid?

      grp.erase! unless grps.empty?
      return grps
    end

    def delete_ents_behind_plane(ents, t1, plane)
      to_delete = []
      ents.each { |e|
        if e.is_a?(Sketchup::Face) && e.valid?
          mesh = e.mesh
          mesh.transform!(t1) # transform face to global coordinates
          p1 = mesh.point_at(1)
          p2 = mesh.point_at(2)
          p3 = mesh.point_at(3)
          cent = [(p1.x+p2.x+p3.x)/3.0, (p1.y+p2.y+p3.y)/3.0, (p1.z+p2.z+p3.z)/3.0]
          result = plane[0]*cent.x+plane[1]*cent.y+plane[2]*cent.z+plane[3]
          to_delete.push(e) if result < 0.0
        elsif e.is_a?(Sketchup::Edge) && e.valid?
          p1 = e.start.position.transform(t1)
          p2 = e.end.position.transform(t1)
          cent = [(p1.x+p2.x)/2.0, (p1.y+p2.y)/2.0, (p1.z+p2.z)/2.0]
          result = plane[0]*cent.x+plane[1]*cent.y+plane[2]*cent.z+plane[3]
          to_delete << e if result < 0.0 and cent.distance_to_plane(plane) > 0.001
          # to_delete << e if cent.distance_to_plane(plane) > 0.1
        end
      }
      ents.erase_entities(to_delete) # delete entities behind the plane
    end

    def get_definition(e)
      return e.entities[0].parent if e.is_a?(Sketchup::Group)
      return e.definition if e.is_a?(Sketchup::ComponentInstance)
      nil
    end

  end # class << self
end # module AMS::SplitToPieces


unless file_loaded?(__FILE__)
  file_loaded(__FILE__)

  UI.menu('Plugins').add_item('Split to Pieces'){
    mod = Sketchup.active_model
    sel = mod.selection
    grps = []
    sel.each { |e|
      grps << e if (e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance))
    }
    if grps.empty?
      UI.messagebox('No selection. At least one group/component must be selected!')
      next
    end
    res = UI.inputbox(['Recurse'], [2], ['0|1|2|3|4|5|6|7'], 'Split Options')
    recurse = res[0].to_i
    if Sketchup.version.to_i > 6
      mod.start_operation('split to pieces', true)
    else
      mod.start_operation('split to pieces')
    end
    grps.each { |e| sel.add AMS::SplitToPieces.split(e, recurse) }
    mod.commit_operation
  }

end
